use std::mem::swap;
use std::ops::Fn;
use std::thread;
use std::env;


extern crate rand;
use rand::{ThreadRng,Rng};


type Element = [u64; 64];
type Slice = u64;

fn init_vec<T, F : Fn() -> T>(f : F, s : usize) -> Vec<T> {
    (0..s).map(|_| f()).collect()
}

fn get_slice(e : &Element, l : usize) -> Slice{
    let mut s = 0;
    let mask = 1 << l;
    for c in 0..e.len() {
        if (e[c] & mask) != 0 {
        s+= 1 << c
        }
    }
    s
}

fn level(e : &Element, w : usize) -> usize {
  for i in 0..w {
    if get_slice(e, i) != 0 { return i }
  }
  w
}

fn add(mut e1 : Element, e2 : Element) -> Element{
  e1.iter_mut().zip(e2.into_iter()).map(|(a,b)| {*a = a.wrapping_add(*b)}).collect::<Vec<()>>();
  e1
}

fn sub(mut e1 : Element, e2 : Element) -> Element{
  e1.iter_mut().zip(e2.into_iter()).map(|(a,b)| {*a = a.wrapping_sub(*b)}).collect::<Vec<()>>();
  e1
}

fn add_sub(e1 : Element, e2 : Element, g : &mut Rnd) -> Element{
  if random_bool(g) {
    add(e1,e2)
  } else {
    sub(e1,e2)
  }
}

fn get_bool(e : &Element, pos : usize, p : usize) -> bool{
  e[pos % p] & (1 << (pos/p)) !=0
}

fn first_one(e: &Element, p : usize, w : usize) -> usize {
  for i in 0..(w*p) {
	if get_bool(e,i,p) { return i }
  }
  w*p
}

fn common(e1 : &Element, e2 : &Element, p : usize, w : usize) -> usize {
  first_one(&sub(e1.clone(), e2.clone()), p, w)
}

fn normalize(e: &mut Element, p : usize, w : usize) {
  let mut neg = false;
  for i in 0..((w-1)*p) {
	if get_bool(e,i,p) {
		if get_bool(e,i+p,p) {
		  neg = true;
		}
		break;
	}
  }
  if neg {
	for i in 0..p {
		e[i] = 0u64.wrapping_sub(e[i])
	}
  }
}

fn reduce(v : Vec<Element>, g : &mut Rnd) -> Element {
  let mut a = v.into_iter();
  let e = a.next().unwrap();
  a.fold(e, |x,y| { add_sub(x,y,g) })
}

struct System { s : Vec<Option<(Slice, Slice, Element)>> }

impl System {

    fn new(p: usize) -> Self {
       System { s : init_vec(|| None, p) }
    }

    fn is_full(&self) -> bool {
      self.s.iter().all(Option::is_some)
    }
    
    // Try to populate a basis. Return a zero-sum list index otherwise
    fn try_insert(&mut self, e : Element, l : usize) -> Option<(Slice, Element)>{
        let mut slice = get_slice(&e, l);
        let mut eqn = 0;
        let ref mut s = self.s;
        for i in 0..s.len() {
            match (&mut s[i], slice & (1 << i) != 0) {
            (ref mut a @ &mut None, true) => {**a = Some((eqn | (1 << i), slice, e)); return None}
            (_, false) => {}
            (&mut Some((eq2, slice2, _)), true) => { eqn ^=eq2; slice^=slice2 }
            }
        }
        assert!(slice == 0);
        Some((eqn, e))
    }
    
    fn to_list(self) -> Vec<Element> {
      self.s.into_iter().filter_map(|x| x.map_or(None,|x| Some(x.2))).collect()
    }
    
    // Get a list of elements from the index list
    fn get_zero_sum_list(&mut self, e: Element, l : usize) -> Option<Vec<Element>>{
        if let Some((eqn, e)) = self.try_insert(e, l){
            //println!("b : {:?}, {}", self.s, eqn);
            let mut elts = vec![e];
            let mut to_reinsert = Vec::new();
            let mut invalid = eqn;
            for i in 0.. self.s.len() {
                if eqn & (1 << i) != 0 {
                    let mut a = None;
                    swap(&mut a, &mut self.s[i]);
                    let (_,_,e) = a.unwrap();
                    elts.push(e);
                } else {
                    if let Some((eq, _, _)) = self.s[i] {
                        if (invalid & eq) != 0 {
                            invalid |= eq;
                            let mut a = None;
                            swap(&mut a, &mut self.s[i]);
                            let (_,_,e) = a.unwrap();
                            to_reinsert.push(e);
                        }
                    }
                }
            }
            for e in to_reinsert.into_iter() {
                assert!(self.try_insert(e,l).is_none())
            }
            //println!("e: {:?}", self.s);
            Some(elts)
        } else { None }
    }

}


struct Rnd { state : u32, lstate : u32, g : ThreadRng }

fn new_rnd() -> Rnd {
  Rnd { state : 0, lstate : 0, g : rand::thread_rng() }
}

fn random_bool(g : &mut Rnd) -> bool {
    if g.lstate == 0 {
      g.lstate = 31;
      g.state = g.g.next_u32();
    } else {
      g.lstate-=1;
    }
    let b = g.state % 2 == 1;
    g.state/=2;
    b
}


fn make_elt(w : usize, p : usize, g : &mut Rnd) -> Element {
  let mut e : Element = [ 0; 64 ];
  for i in 0..p {
    e[i] = g.g.next_u64()
  }
  e
}

fn drop_basis(w : &mut Vec<Element>, l : usize, p : usize) -> bool {
  let mut s = System::new(p);
  let mut a = true;
  let mut v = Vec::new();
  swap(&mut v, w);
  *w = v.into_iter().filter_map( |e| 
    {if a { a=!s.is_full();
      s.try_insert(e,l).map(|x| x.1)}
      else {Some(e)}
    }).collect();
  s.is_full()
}

fn sieve(list : &mut Vec<Element>, trees : &mut Vec<Tree>, elts : &mut Vec<Vec<Element>>, l: usize, p : usize, w : usize, g : &mut Rnd) {
  let mut s = System::new(p);
  let mut v = Vec::new();
  swap(&mut v, list);
  v.into_iter().filter_map( |e| {s.get_zero_sum_list(e,l)}).map(|v| {let e = reduce(v,g); insert_elt(e,trees, elts, l, p, w)}).last();
}

fn insert(list : &mut Vec<Vec<Element>>, e : Element, w : usize) {
  let l = level(&e,w-1);
  list[l].push(e)
}

fn usage(s : &str) {
 println!("Usage : {} P W SAMPLES #ITERATIONS #PROCESS THRESHOLD", s); 
}

//#[derive(Debug)]
enum Tree{
  Empty,
  Leaf { nb : i32, leaf : Element },
  Node { len : i32, best : bool, lchild : Box<Tree>, rchild : Box<Tree>, nb : usize }
}

use Tree::*;


//Create the needed nodes for a new element
fn new_node(c: usize, t : &mut Tree, r : Element, pos : usize, p : usize, nb : i32){
  let b = get_bool(&r,c,p);
  let common  = r.clone();
  let new_leaf = Leaf {nb : 1, leaf : r};
  let nbs = nb as usize +1;
  let mut old_leaf = Empty;
  swap(t, &mut old_leaf);
  let (l,r) = if b {(old_leaf,new_leaf)} else {(new_leaf,old_leaf)};
  let mut tree = Node { len : 0, best : false, lchild : Box::new(l), rchild : Box::new(r) , nb : nbs};
  for i in 1..(c + 1 - pos) {
    let e = get_bool(&common,c - i,p);
    let (lchild, rchild) = if e {(Box::new(Empty), Box::new(tree))} else {(Box::new(tree), Box::new(Empty))};
    tree = Node { len : i as i32, best : e, lchild : lchild, rchild : rchild, nb : nbs };
  }
  swap(t, &mut tree)
}

// Gives the length of the best collision in the tree, -1 means no collision
fn length(t : & Tree) -> i32{
  match *t {
    Empty | Leaf { .. } => -1,
    Node { len : t, .. } => t
  }
}

// Insert recursively a new element in the tree
fn insert_tree(e: Element, tree: &mut Tree, pos: usize, p : usize, w : usize, n: usize) -> i32{
    match tree {
        t @ &mut Empty => { *t = Leaf{ nb : 1, leaf : e }; -1 },
        t @ &mut Leaf{ .. } => { 
          let (c, nb) = match t {
            &mut Leaf{ ref mut leaf, ref mut nb } => { let g = common(&leaf, &e, p, w); if g >= n {*nb+=1; return -1}; (g, *nb) },
            _ => unreachable!()
          };
          new_node(c, t , e, pos, p, nb); 
          (c - pos) as i32
        }
        &mut Tree::Node { ref mut len, ref mut best, ref mut lchild, ref mut rchild, ref mut nb } => {
            *nb+=1;
            if get_bool(&e,pos,p) {
                let l2 = insert_tree(e, rchild, pos+1, p, w, n);
                if l2 >= *len{
                        *best = true;
                        *len = l2+1;
                }
            } else {
                let l2 = insert_tree(e, lchild, pos+1, p, w, n);
                if l2 >= *len{
                    *best = false;
                    *len = l2+1;
                }
            }
            *len
        }
    }
}


// Normalize the element, then insert it
fn canonical_insert(mut e: Element, tree: &mut Tree, pos: usize, p : usize, w : usize){
    //normalize(&mut e,p,w);
    insert_tree(e, tree, pos, p, w, p*w);
}

// Simplify the tree if necessary after popping an element
fn sanitize(t : &mut Tree){
 let mut garbage = Empty;
 match t {
   &mut Leaf { nb : 0, .. } => { },
   &mut Node { ref mut lchild, ref mut rchild, .. } => {
     match (lchild.as_mut(), rchild.as_mut()) {
       (&mut Empty, &mut Empty) => { }
       (&mut Empty, r @ &mut Leaf { .. }) => { swap(&mut garbage, r) },
       (l @ &mut Leaf { .. }, &mut Empty) => { swap(&mut garbage, l) },
       _ => { return }
       }
   },
   _ => { return }
 };
 swap(t, &mut garbage)
}


// Pop the two elements with the best potential collision
fn pop_best(t : &mut Tree) -> (Element, Element){
  let (e1,e2) = match t {
    &mut Empty | &mut Leaf { nb : 1, .. } => { panic!("Pop'd in an empty tree") },
    &mut Leaf { ref mut nb, leaf : ref l } => { *nb-=2; (l.clone(), l.clone()) },
    &mut Node { ref mut len, ref mut best, ref mut lchild, ref mut rchild, ref mut nb } => {
      *nb-=2;
      let (e1, e2, tosan) = match (lchild.as_mut(), rchild.as_mut()) {
        (&mut Leaf { nb : ref mut nbl, leaf : ref l }, &mut Leaf { nb : ref mut nbr, leaf : ref r }) =>
          { *nbl-=1; *nbr-=1; (l.clone(), r.clone(), true) },
        (lchild @ _, rchild @ _) => {
          let (e1, e2) = if *best {pop_best(rchild)} else {pop_best(lchild)};
          let (lleft, lright) = (length(lchild), length(rchild));
          if lleft > lright {
            *len = lleft+1;
            *best = false;
          } else {
            *len = lright+1;
            *best = true;
          }
        (e1, e2, false)
        }
      };
     if tosan {sanitize(lchild); sanitize(rchild)};
     (e1, e2)
     }
  };
  sanitize(t);
  (e1,e2)
}

fn drop_tree(t : & Tree, v : &mut Vec<Element>){
  match *t {
    Empty => {},
    Leaf {nb, ref leaf} => {for _ in 0..nb {v.push(leaf.clone())} }
    Node {  len, best, ref lchild, ref rchild, nb } => {
	drop_tree(lchild, v);
        drop_tree(rchild, v);
    }
  }
}
fn nb_elt( t: & Tree) -> i32 {
  match *t {
    Node { ref lchild, ref rchild, .. } => { nb_elt(lchild) + nb_elt(rchild) }
    Leaf { nb, .. } => {nb}
    Empty => {0}
  }

}

fn insert_trees(l : &mut Vec<Tree>, e : Element, p : usize, w : usize) {
	let o = first_one(&e,p,w);
	if o < w * p {
		canonical_insert(e, &mut l[o%p], o, p, w)
	}
}

fn insert_elt(e : Element, list_tree : &mut Vec<Tree>, list_elts : &mut Vec<Vec<Element>>, l : usize, p : usize, w : usize) {
             if level(&e, w) == l {
                insert_trees(list_tree,e,p,w)
              } else {
                insert(list_elts, e, w);
	      }
 
}

fn print_vec_vec(v : &Vec<Vec<i32>>) {
  for i in 0..v.len() {
    println!("{}, {}\t: {:?}", i, v[i].iter().enumerate().max_by_key(|x| x.1).unwrap().0, v[i]);
  }

}
fn compute(p : usize, w : usize, s: usize, i : usize, t : usize) ->  (usize, Vec<usize>){
  let mut ok = 0usize;
  let mut num = vec![0; w];
  let mut g = new_rnd();
  for _ in 0..i {
    let mut cont = true;

    let mut list_elts = init_vec(|| Vec::new(), w);
    for _ in 0..s {
	insert(&mut list_elts, make_elt(w,p, &mut g), w)
    }
    for l in 0..w {
      let mut list_tree = init_vec(|| Tree::Empty,p);
      let mut v = Vec::new();
      let mut nbs = init_vec(|| vec![0 ;p],p);
      swap(&mut v, &mut list_elts[l]);
      num[l]+=v.len();
      num[l]+=(l+1..w).map(|x| {list_elts[x].len()}).sum::<usize>();
      cont = drop_basis(&mut v, l, p) && cont;
      //if ! cont {println!("Failed at level {}", l);};
      let size = v.len();
      let mut tree = Empty;
      for e in v.into_iter() {
        insert_trees(&mut list_tree,e,p,w)
      }
      let mut v = Vec::new();
      
      for m in 0..p {
          let mut x = 0;
          let mut y = 0;
          while length(&list_tree[m]) >= 0 && l < t {
	    let (a,b) = pop_best(&mut list_tree[m]);
	    let c = add_sub(a,b,&mut g);
	    let a = first_one(&c,p,w);y+=1;
	    x+=a;
	    if a < p { nbs[m][a]+=1 }
            insert_elt(c, &mut list_tree, &mut list_elts, l, p, w);
          }
          //if l < t {println!("gain pour {:?} : {}", (m,l), x as f64/y as f64 - ((m+l*p) as f64))};
          drop_tree(&list_tree[m], &mut v);
      }
      //if l == 0 {print_vec_vec(&nbs)}
      g.g.shuffle(&mut v);
      sieve(&mut v, &mut list_tree, &mut list_elts, l, p, w, &mut g)
    }
    if cont { ok+=1; };
  }
  (ok,num)
}

fn main() {
    let args : Vec<String> = env::args().collect();
    if args.len() != 7 { return usage(&args[0]) };
    let (p, w, s, i, c, t) : (_,_,usize, usize, usize, _) = match (args[1].parse(),args[2].parse(),args[3].parse(),args[4].parse(),args[5].parse(), args[6].parse()) {
      (Ok(p), Ok(w), Ok(s), Ok(i), Ok(c), Ok(t)) => {(p,w,s,i,c,t)},
      _ => { return usage(&args[0]) }
    };
    let mut threads = Vec::with_capacity(c-1);
    let  a = i / c;
    let b = i % c;
    for j in 0..c-1 {
      let i = if j < b { a+1 } else { a };
      threads.push(thread::spawn( move || compute(p,w,s,i,t) ))
    }
    let mut samples = a;
    let (mut ok, mut vec) = compute(p,w,s,a,t);
    let mut j = 0;
    for t in threads {
      match t.join() {
        Ok((r,v)) => { ok+= r; vec.iter_mut().zip(v.iter()).map(|(x,y)| *x+=*y).collect::<Vec<()>>();samples += if j < b { a+1 } else { a }; },
        Err(e) => { println!("Failed : {:?}",e)}
      }
      j+=1;
    }
    println!("{}", ok as f64 / samples as f64);
    for i in 0..w {
       println!("Level {} : {}", i, vec[i] as f64 / samples as f64);
    }

}
